#include <util/heap.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

using Util::Heap;

struct Data {
	unsigned int magic;
	unsigned int data;
	Heap<Data>::Node heap_node;
	
	bool operator < (Data &d)
	{
		return data < d.data;
	}
	
	Data()
	{
		magic = 0xdeadbeef;
	}
};

static const int heap_size = 10000;

Data data[heap_size];

int main(void)
{
	Heap<Data> heap;
	srandom(1);
	
	for (int i = 0; i < heap_size; i++) {
		Data *d = &data[i];
		d->data = random() & 0x7fffffff;
		heap.add(d);
	}
	
	unsigned int last = 0;
	
	for (int i = 0; i < heap_size; i++) {
		Data *d = heap.get_top();
		if (d == NULL) {
			printf("FAILED: returned NULL early\n");
			exit(1);
		}
		
		if (d->magic != 0xdeadbeef) {
			printf("FAILED: garbage data returned\n");
			exit(1);
		}
		
		if (d->data >= 0x80000000) {
			printf("FAILED: duplicate or garbage data returned\n");
			exit(1);
		}
		
		heap.del(d);
		
		if (d->data < last) {
			printf("FAILED: not monotonically increasing\n");
			exit(1);
		}
		
		last = d->data;
		d->data = 0x80000000;
	}

	for (int i = 0; i < heap_size; i++) {
		Data *d = &data[i];
		d->data = random() & 0x7fffffff;
		heap.add(d);
	}
	
	last = 0;
	
	for (int i = 0; i < heap_size; i++) {
		Data *d = heap.get_top();
		if (d == NULL) {
			printf("FAILED: with requeue: returned NULL early\n");
			exit(1);
		}
		
		if (d->magic != 0xdeadbeef) {
			printf("FAILED: with requeue: garbage data returned\n");
			exit(1);
		}
		
		if (d->data >= 0x80000000) {
			printf("FAILED: with requeue: duplicate or garbage data returned\n");
			exit(1);
		}
		
		heap.del(d);
		
		if (d->data < last) {
			printf("FAILED: with requeue: not monotonically increasing\n");
			exit(1);
		}
		
		last = d->data;
		d->data = 0x80000000;
		
		for (int j = 0; j < 10; j++) {
			d = &data[random() % heap_size];
			
			if (d->data == 0x80000000)
				continue;
			
			d->data = random() & 0x7fffffff;
			heap.requeue(d);
			
			if (d->data < last)
				last = d->data;
		}
	}
	
	printf("PASSED\n");
}
