// arch/x64/thread.cc -- Thread switching
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/thread.h>
#include <kern/mem.h>
#include <kern/pagetable.h>

namespace Arch {
	void switch_thread(Threads::Thread *dest, Threads::Thread *src)
	{
		u64 dummy1, dummy2;
		
		if (dest->addr_space) {
			assert(dest->addr_space == dest->active_addr_space);
		
			if (dest->addr_space != src->active_addr_space) {
				u64 cr3 = Mem::kvirt_to_phys(dest->addr_space->
				                             page_table->toplevel);
				asm volatile("movq %0, %%cr3" : : "r" (cr3) : "memory");
			}
		} else {
			dest->active_addr_space = src->active_addr_space;
		}
		
		Priv::tss.rsp[0] = reinterpret_cast<u64>(dest);

		asm volatile("movq %%rsp, (%0);"
		             "movq %%rbp, 8(%0);"
		             "movb $0, 16(%0);"
		             "movb 16(%1), %%al;"
		             "cmpb $0, %%al;"
		             "movq (%1), %%rsp;"
		             "movq 8(%1), %%rbp;"
		             "jnz x64_new_thread;" :
		             "=a" (dummy1), "=c" (dummy2) :
		             "0" (&src->arch.rsp), "1" (&dest->arch.rsp) :
		             "rbx", "rdx", "rsi", "rdi",
		             "r8", "r9", "r10", "r11", "r12", 
		             "r13", "r14", "r15", "memory");
	}
}
