#ifndef _ARCH_MEM_H
#define _ARCH_MEM_H

#include <kern/types.h>
#include <arch/paging.h>

extern int _end;

namespace Mem {
	class PageAllocZone;
};

namespace Arch {
	// First and last+1 pages of memory, as page numbers.  Small holes are
	// ignored; large holes will require discontiguous memory support.
	
	static const size_t mem_start = 0;
	extern size_t mem_end;
	
	static const uintptr_t highzonestart = 4UL*1024*1024*1024 / page_size;
	static const uintptr_t dma32zonestart = 16*1024*1024 / page_size;

	extern Mem::PageAllocZone *pagezones[];
	extern Mem::PageAllocZone **pagezonelists[3];

	extern uintptr_t next_free_bootmem;
	
	namespace Priv {
		struct Descriptor {
			u16 limit_low;
			u16 base_low;
			u8 base_mid;
			u8 type:4;
			u8 user:1;
			u8 dpl:2;
			u8 present:1;
			u8 limit_high:4;
			u8 sw:1;
			u8 code64:1;
			u8 opsize:1;
			u8 gran:1;
			u8 base_high;
		};

		extern Mem::PageAllocZone isadmazone, dma32zone, highzone;

		void early_adjust_mappings();

		// Call after mem_end is set.
		void map_physmem();
	}
}

extern Arch::Priv::Descriptor x64_gdt[1024];

#endif
