#ifndef _ARCH_X86C_MULTIBOOT_H
#define _ARCH_X86C_MULTIBOOT_H

#include <kern/types.h>

namespace Arch {
namespace Priv {
namespace MultiBoot {
	struct BootInfo {
		enum {
			flag_mem,
			flag_boot_dev,
			flag_cmdline,
			flag_modules,
			flag_aout_syms,
			flag_elf_syms,
			flag_mmap,
			flag_drives,
			flag_config,
			flag_bootloader_name,
			flag_apm,
			flag_vbe
		};
	
		u32 flags;
		u32 mem_lower, mem_upper;
		u32 boot_device;
		u32 cmdline;
		u32 mods_count, mods_addr;
		union {
			struct {
				u32 tabsize;
				u32 strsize;
				u32 addr;
				u32 reserved;
			} aout;
			struct {
				u32 num, size, addr, shndx;
			} elf;
		};
		u32 mmap_length, mmap_addr;
	};
		
	struct Module {
		u32 start, end;
		u32 str;
		u32 reserved;
	};
	
	struct MemMap {
		u32 size;
		u64 base;
		u64 len;

		// Possible values for type:
		enum {
			Available = 1,
			Reserved,
			ACPIReclaim,
			ACPINVS
		};

		u32 type;
	} __attribute__((packed));
	
	void process_info();
}
}
}

#endif
