// orb/invoke.cc -- Method Invocation
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/types.h>
#include <kern/orb.h>
#include <kern/pagealloc.h>
#include <kern/compiler.h>
#include <kern/thread.h>

#include <arch/usercopy.h>

#include <orb.h>

using System::RunTime::ParamInfoBlock;
using namespace ORB;
using Threads::Thread;

namespace ORB {
	static CallFrame *new_frame(Thread *thread)
	{
		CallStackHeader *hdr = thread->orbstack_top_hdr;
		
		if (unlikely(thread->orbstack_top == hdr->num_frames - 1)) {
			hdr = static_cast<CallStackHeader *>(Mem::alloc_pages(1));
			new(hdr) CallStackHeader;
			hdr->thread = thread;
			hdr->num_frames = (Arch::page_size - sizeof(CallStackHeader)) /
			                  sizeof(CallFrame);

			thread->orbstack_top_hdr->node.add_front(&hdr->node);
			thread->orbstack_top_hdr = hdr;
			thread->orbstack_top = 0;
			return &hdr->frames[0];
		}
		
		return &hdr->frames[++thread->orbstack_top];
	}

	ObjectHdr *IDSpace::get_local_hdr(Object *obj)
	{
		ObjectHdr *hdr = *rmap.lookup(obj->id);
		if (hdr)
			return hdr;
		
		int id = alloc.alloc();
		ObjectPtr *ptr = &table.lookup(id, true)->ptr;

		ptr->id = id;
		ptr->flags = 0;
		ptr->Pointer = 1;
		ptr->object = obj;

		*rmap.lookup(id, true) = ptr;
		return ptr;
	}

	Object *IDSpace::newobj(Mem::ProcAddrSpace *aspace, uintptr_t entry)
	{
		int id = alloc.alloc();
		Object *obj = &table.lookup(id, true)->obj;

		obj->id = id;
		obj->flags = 0;
		obj->aspace = aspace;
		obj->entry = entry;
		
		return obj;
	}

	void invoke_method(ParamInfoBlock *user_pib, uintptr_t &stack)
	{
		ParamInfoBlock pib = Arch::copyin(user_pib);
		CallFrame *frame = new_frame(curthread);
		
//		if (pib->objlist_len == 0)
			
		
//		frame->object = objid;
		frame->caller_user_pib = user_pib;
//		frame->ret_pc = ret_pc;
		
		printf("invoke_method: frame %p pib %p ret %lx\n",
		       frame, frame->caller_user_pib,
		       frame->ret_pc);
		
		
	}
	
	uintptr_t return_from_method(uintptr_t &exptr, size_t &exlen)
	{
		return 0;
	}
}
